#include<math.h>
#include<stdlib.h>
#include<stdio.h>

//#define INTERIOR       0
//#define FRONTERA_REAL  1

#include "libs/jag_mem.cpp"
#include "libs/jag_phi.cpp"
#include "libs/jag_vector.cpp"
#include "libs/jag_cond.cpp"
#include "libs/qSort.cpp"

#include "edp_ut.cpp"

#include "libs/tree2d.cpp"

#include "libs/interpola_tps2d.cpp"

#include "varios.cpp"
#include "varios1.cpp"


void reconstruye(double *x,double *y, double *lambda,double c,int N, double *sol_numerica);
double interpola(double xc,double yc, double *x,double *y, double *lambda,double c,int N);
void mult_matrix_vec(double **P, double *lambda,int m,double *z);
void corrige_bc(double *x,double *y, int ni, int m, double t,double *f);
double* resuelve_IC(double *x,double *y, int N);
void crea_P(double *x,double *y, int N,double stept,double **P);
void crea_M(double *x,double *y,int ni, int N,double stept,double **M);

//--------------------------------------------------------------------
void reconstruye(double *x,double *y, double *lambda,double c,int N, double *sol_numerica)
{
int     i,j;
double xc,yc;
double s;

 for(i=0;i<N;i++)
 {
  s=0.0;
   for(j=0;j<N;j++)
   {
      s += lambda[j]*phitps42(x[i]-x[j],y[i]-y[j],0,0);
   }

   s = s + lambda[N+0];
   s = s + lambda[N+1]*x[i];
   s = s + lambda[N+2]*y[i];
   s = s + lambda[N+3]*x[i]*y[i];
   s = s + lambda[N+4]*x[i]*x[i];
   s = s + lambda[N+5]*y[i]*y[i];

   sol_numerica[i] = s;
 }
}

//--------------------------------------------------------------------
double interpola(double xc,double yc, double *x,double *y, double *lambda,double c,int N)
{
 int      j;
double s;

  s=0.0;
   for(j=0;j<N;j++)
   {
      s += lambda[j]*phitps42(xc-x[j],yc-y[j],0,0);
   }

   s = s + lambda[N+0];
   s = s + lambda[N+1]*xc;
   s = s + lambda[N+2]*yc;
   s = s + lambda[N+3]*xc*yc;
   s = s + lambda[N+4]*xc*xc;
   s = s + lambda[N+5]*yc*yc;

 return s;
 
}

//--------------------------------------------------------------------
//Proyeccion
void mult_matrix_vec(double **P, double *lambda,int m,double *z)
{
 double s;	
  for(int i=0;i<m;i++)	
  {
	  s = 0.0; 
   for(int j=0;j<m;j++)	
   {
	 s = s + P[i][j]*lambda[j];
   }	     
  s = s + P[i][m+0]*lambda[m+0];
  s = s + P[i][m+1]*lambda[m+1];
  s = s + P[i][m+2]*lambda[m+2];
  s = s + P[i][m+3]*lambda[m+3];
  s = s + P[i][m+4]*lambda[m+4];
  s = s + P[i][m+5]*lambda[m+5];

  z[i] = s; 
 } 
      
}
//--------------------------------------------------------------------
//Correcion
void corrige_bc(double *x,double *y, int ni, int m, double t,double *f)
{
  for(int i=ni;i<m;i++)		
     f[i] = u_t(x[i],y[i],t);
}
//--------------------------------------------------------------------
double* resuelve_IC(double *x,double *y, int N)
{
 double **A;
 double  *f;
 double  *lambda;
 int     *pivot_index;
 double   cond;
 int      Pol=6;

//Reservo la memoria dinamica
  A           = make_dmatrix(N+Pol,N+Pol);
  f           = make_dvector(N+Pol);
  lambda      = make_dvector(N+Pol);
  pivot_index = make_ivector(N+Pol);

//Relleno de zeros
  fill(f,N+Pol,0.0);

 for(int i=0;i<(N+Pol);i++)
   fill(A[i],N+Pol,0.0);

//Creo la matriz de interpolacion  
for(int i=0;i<N;i++)
{
    for(int j=0;j<N;j++)
    {
        A[i][j] = phitps42(x[i]-x[j],y[i]-y[j],0,0);
    }
  f[i]=u_t0(x[i],y[i]); 

 A[i][N+0] = 1.0;
 A[i][N+1] = x[i];
 A[i][N+2] = y[i];
 A[i][N+3] = x[i]*y[i];
 A[i][N+4] = x[i]*x[i];
 A[i][N+5] = y[i]*y[i];

 A[N+0][i] = 1.0;
 A[N+1][i] = x[i];
 A[N+2][i] = y[i];
 A[N+3][i] = x[i]*y[i];
 A[N+4][i] = x[i]*x[i];
 A[N+5][i] = y[i]*y[i];

}

//Factorizo la matriz
   Factor(A,N+Pol,&cond,pivot_index);

   printf("num.condi   = %e\n",cond);  

//Resuelvo mediante LU
  copy(f,N+Pol,lambda);
  Solve(A,N+Pol,pivot_index,lambda);

//Libero la memoria dinamica
  free_dmatrix(A,N+Pol);
  free_dvector(f);
  free_ivector(pivot_index);

 return lambda;
}

//-------------------------------------------------------------------------------------------------------
//Creo la matriz-P
void crea_P(double *x,double *y, int N,double stept,double **P)
 {
double u,ux,uy;

  for(int i=0;i<N;i++)
  {
   for(int j=0;j<N;j++)
   {
	  u  = phitps42(x[i]-x[j],y[i]-y[j],0,0);   
	 ux  = phitps42(x[i]-x[j],y[i]-y[j],0,1);   
	 uy  = phitps42(x[i]-x[j],y[i]-y[j],0,3);   	 
     P[i][j] = u - 0.5*stept*(ux+uy);
   } 
    P[i][N+0] = 1.0       - 0.5*stept*0.0      - 0.5*stept*0.0;
    P[i][N+1] = x[i]      - 0.5*stept*1.0      - 0.5*stept*0.0;
    P[i][N+2] = y[i]      - 0.5*stept*0.0      - 0.5*stept*1.0;
    P[i][N+3] = x[i]*y[i] - 0.5*stept*y[i]     - 0.5*stept*x[i];
    P[i][N+4] = x[i]*x[i] - 0.5*stept*2.0*x[i] - 0.5*stept*0.0;
    P[i][N+5] = y[i]*y[i] - 0.5*stept*0.0      - 0.5*stept*2.0*y[i];
  }
}

//-------------------------------------------------------------------------------------------------------
//Creo la matriz-M
void crea_M(double *x,double *y,int ni, int N,double stept,double **M)
 {
 double u,ux,uy;

 for(int i=0;i<(N+6);i++)
   fill(M[i],N+6,0.0);

  for(int i=0;i<ni;i++)
  {
   for(int j=0;j<N;j++)
   {
	  u = phitps42(x[i]-x[j],y[i]-y[j],0,0);   
	 ux = phitps42(x[i]-x[j],y[i]-y[j],0,1);   
	 uy = phitps42(x[i]-x[j],y[i]-y[j],0,3);   	 
     M[i][j] = u + 0.5*stept*(ux+uy);
   } 
    M[i][N+0] = 1.0  + 0.5*stept*0.0 + 0.5*stept*0.0;
    M[i][N+1] = x[i] + 0.5*stept*1.0 + 0.5*stept*0.0;
    M[i][N+2] = y[i] + 0.5*stept*0.0 + 0.5*stept*1.0;
    M[i][N+3] = x[i]*y[i] + 0.5*stept*y[i] + 0.5*stept*x[i];
    M[i][N+4] = x[i]*x[i] + 0.5*stept*2.0*x[i] + 0.5*stept*0.0;
    M[i][N+5] = y[i]*y[i] + 0.5*stept*0.0 + 0.5*stept*2.0*y[i];

    M[N+0][i] = 1.0;
    M[N+1][i] = x[i];
    M[N+2][i] = y[i];
    M[N+3][i] = x[i]*y[i];
    M[N+4][i] = x[i]*x[i];
    M[N+5][i] = y[i]*y[i];
  }

  for(int i=ni;i<N;i++)
  {
   for(int j=0;j<N;j++)
   {
	  u = phitps42(x[i]-x[j],y[i]-y[j],0,0);   
     M[i][j] = u;
   } 
    M[i][N+0] = 1.0;
    M[i][N+1] = x[i];
    M[i][N+2] = y[i];
    M[i][N+3] = x[i]*y[i];
    M[i][N+4] = x[i]*x[i];
    M[i][N+5] = y[i]*y[i];

    M[N+0][i] = 1.0;
    M[N+1][i] = x[i];
    M[N+2][i] = y[i];
    M[N+3][i] = x[i]*y[i];
    M[N+4][i] = x[i]*x[i];
    M[N+5][i] = y[i]*y[i];
  }
}
//-------------------------------------------------------------------------------------------------------
//Creo la matriz-A
void crea_A(double *x,double *y, int N,double **A,double *f)
 {
//Relleno de zeros
  fill(f,N+6,0.0);

  for(int i=0;i<(N+6);i++)
    fill(A[i],N+6,0.0);


//Creo la matriz de interpolacion  
for(int i=0;i<N;i++)
{
    for(int j=0;j<N;j++)
    {
        A[i][j] = phitps42(x[i]-x[j],y[i]-y[j],0,0);
    }
  f[i]=u_t0(x[i],y[i]); 

 A[i][N+0] = 1.0;
 A[i][N+1] = x[i];
 A[i][N+2] = y[i];
 A[i][N+3] = x[i]*y[i];
 A[i][N+4] = x[i]*x[i];
 A[i][N+5] = y[i]*y[i];

 A[N+0][i] = 1.0;
 A[N+1][i] = x[i];
 A[N+2][i] = y[i];
 A[N+3][i] = x[i]*y[i];
 A[N+4][i] = x[i]*x[i];
 A[N+5][i] = y[i]*y[i];
}
}
//-----------------------------------------------------------------------------------------------------
void refina_initial_condition(tree *raiz, double thetar, double thetac,int min_nivel_ref, int max_nivel_ref)
{
  int    ni,N;
  int    band;
  double *x,*y; 
  double  *lambdaIC;
  tree      **data;
  double  *sol_numerica;
  double  *sol_real;
  double  maxi_elocal;
   int ref,rem;
  int restriccion;

band = 0;
for(int iter=0;iter<50;iter++)
{
   clean_frontera(raiz);
   marca_frontera(raiz);

// Ajusto las fronteras
 //  ajusta_fronteras(raiz,xmin,xmax,ymin,ymax);

//Obtengo los nodos tipo frontera e interiores
   N=0; cont_hojas(raiz,N);

   printf("------------------------------------ iter=%d, N=%d\n",iter,N);

   x       = make_dvector(N); 
   y       = make_dvector(N);
   data = (tree **)malloc(N*sizeof(tree));   
   sol_numerica = make_dvector(N);
  

   ni=0; N=0;
   get_datos_interior(raiz,x,y,data,ni);
   N=ni;
   get_datos_frontera(raiz,x,y,data,N);
   
   printf("ni=%d    n=%d\n",ni,N);

   lambdaIC = resuelve_IC(x,y,N);

   reconstruye(x,y,lambdaIC,0,N,sol_numerica);

   for(int i=0;i<N;i++)
     data[i]->sol = sol_numerica[i];

   get_error_localIF(data,N,maxi_elocal);
 
 //Refino y remuevo
  if(maxi_elocal<=thetar)
  {
    printf("-----------------------------------------------------------\n");
    printf("Se alcanzo el error de interpolacion local: %f<%f.\n",maxi_elocal,thetar);
    printf("-----------------------------------------------------------\n\n");
    band = 1;
  }
 if(band==0)
 {
  //Refino los condenados nodos
   ref=0;   rem=0;
   refinaIF(raiz,data,N,min_nivel_ref,max_nivel_ref, thetar,thetac,ref,rem);
   printf("Refinados = %d , Removidos = %d,  newN = %d\n",ref,rem,N+3*ref-3*rem);

//Aplico la restriccion de suavidad
    for(int pp=0;pp<4;pp++)
    {
          restriccion=0;         restriccion_suavidad2(raiz,restriccion,max_nivel_ref);
     }
  }//if(band==0)

  if(band==1)
  {
    save_gnu_solucion("numerica",x,y,sol_numerica,N);

    sol_real           = make_dvector(N);

    for(int i=0;i<N;i++)   sol_real[i] = u_t0(x[i],y[i]);

    save_gnu_solucion("real",x,y,sol_real,N);  

    free_dvector(sol_real);

    iter=500;
  }
 
   free_dvector(x); 
   free_dvector(y);
   free_dvector(lambdaIC);
   free(data);
   free_dvector(sol_numerica);
}//iter
}


//--------------------------------------------------------------------
int main(int argc, char **argv)
{
 double  t,tini,tmax,stept;	
 int     n,N,ni;
 double *sol_numerica;
 double *sol_analitica;
 double *sol_analitica_numerica; 
 double *x,*y; 
 double  e_rms,e_inf;
 double xmin,xmax,ymin,ymax;
 int    max_nivel_ref;
 int    min_nivel_ref; 
 int    nivel_ref;
 double maxi_elocal;
 double thetar,thetac;
 tree    *raiz;
 tree   **data;


 if(argc!=7){
   printf("ERROR:  requiere   <min_nivel>  <thetar> <thetac>   <sigma> <stept> <tmax>\n\n");
   exit(1);	 
 }
 
nivel_ref     = atoi(argv[1]);
thetar        = atof(argv[2]);  
thetac        = atof(argv[3]);  
SIGMA     = atof(argv[4]);
stept         = atof(argv[5]);  
tmax          = atof(argv[6]);  

 tini = 0.0;


 max_nivel_ref = 15;
 min_nivel_ref = 3;   // 128 nodos


 printf("-----------------------------------------------------\n");
 printf("nivel_ref = %d, total = %f\n",nivel_ref,pow(4,nivel_ref));
 printf("thetar      = %f\n",thetar);   
 printf("thetac      = %f\n",thetac);    
 printf("sigma      = %f\n",SIGMA);
 printf("stept       = %f\n",stept);   
 printf("tmax        = %f\n",tmax);    
 

//Creo el grid cartesiano
 xmin =  0;
 ymin =  0;
 xmax =  1;
 ymax =  1;

//Creo la raiz inicial del arbol quad-tree
  raiz = crea_raiz(xmin,xmax,ymin,ymax);


//Refino todas las hojas varias veces--jeje
  for(int i=0;i<nivel_ref;i++){
    refina_todo(raiz,max_nivel_ref); clean_band(raiz);
}


//Ciclo Iterativo para Refinar la condicion inicial de la EDP
  refina_initial_condition(raiz,thetar,thetac,min_nivel_ref,max_nivel_ref);

//Ya termine RLN para la condicion inicial, ahora viene la parte temporal
  double **A,**M,**P;
  double *f,*lambda;
  int    *pivot_index;
  double  cond;
  double  u,ux,uy;

//Obtengo los datos
   N=0; cont_hojas(raiz,N);
   x    = make_dvector(N); 
   y    = make_dvector(N);
   data = (tree **)malloc(N*sizeof(tree));   
   ni=0; N=0;   get_datos_interior(raiz,x,y,data,ni);
   N=ni;        get_datos_frontera(raiz,x,y,data,N);

//Reservo la memoria dinamica
  A           = make_dmatrix(N+6,N+6);
  M           = make_dmatrix(N+6,N+6);
  P           = make_dmatrix(N+6,N+6);
  f           = make_dvector(N+6);
  lambda      = make_dvector(N+6);
  pivot_index = make_ivector(N+6);

  sol_analitica          = make_dvector(N);
  sol_numerica           = make_dvector(N);
  sol_analitica_numerica = make_dvector(N);   


//Creo la matriz de interpolacion 
  crea_A(x,y,N,A,f);

//Factorizo la matriz
   Factor(A,N+6,&cond,pivot_index);
   printf("num.condi   = %e\n",cond);  
 
//Resuelvo mediante LU
  copy(f,N+6,lambda);    Solve(A,N+6,pivot_index,lambda);
 

//Calculo el error cuadratico medio y el error maximo
//sobre el mismo grid
  for(int i=0;i<N;i++)
    sol_analitica[i] = u_t0(x[i],y[i]);

  reconstruye(x,y,lambda,0,N,sol_numerica);

  e_rms = calc_erms(sol_analitica,sol_numerica,N);
  e_inf = calc_einf(sol_analitica,sol_numerica,N);
  
  for(int i=0;i<N;i++) 
      sol_analitica_numerica[i] = sol_analitica[i]-sol_numerica[i];     
    
  save_gnu_solucion("numerica",x,y,sol_numerica,N);  
  save_gnu_solucion("real",x,y,sol_analitica,N);     
  save_gnu_solucion("realnumerica",x,y,sol_analitica_numerica,N);      


 printf("t=%f, e_rms       = %e\n",0.0,e_rms);
 printf("t=%f, e_inf          = %e\n",0.0,e_inf); 


//Creo la matriz-P  
  crea_P(x,y,N,stept,P);

//Creo la matriz-M  
 crea_M(x,y,ni, N,stept,M);

//Factorizo la matriz-M
   Factor(M,N+6,&cond,pivot_index);

   printf("num.condi   = %e\n",cond);  
   

//Itero sobre el tiempo   
  t = tini;
  
  int cont=0;

  double *lambda_old = make_dvector(2*N);
  while(t<tmax)
  {
    printf("--------------------------------------------------------------- time = %f , N=%d\n",t,N);
 
     copy(lambda,N+6,lambda_old);

    //Proyeccion
       fill(f,N+6,0.0);                   mult_matrix_vec(P,lambda,N,f);
    //Correcion
       corrige_bc(x,y,ni,N,t+stept,f);      
    //Solucion	 
       copy(f,N+6,lambda);       Solve(M,N+6,pivot_index,lambda);
       
    //Avanzo en el tiempo   
       t = t + stept; 

      cont++;

    //if(!(cont%2))
     { 
     // copy(lambda_old,N+6,lambda);

      //Reconstruyo la solucion 
         reconstruye(x,y,lambda,0,N,sol_numerica);

      //Asigno la solucion 
        for(int i=0;i<N;i++)  data[i]->sol = sol_numerica[i];

 
      //Obtengo el error de interpolacion local 
        get_error_localIF(data,N,maxi_elocal);

         double *tmp = make_dvector(N);
         for(int i=0;i<N;i++)           tmp[i] = data[i]->error;
        save_gnu_solucion("errorlocal",x,y,tmp,N);  
         free_dvector(tmp); 
 

      //Pregunto si se rebaso el error de interpolacion local
       if(maxi_elocal>thetar)
       { 
         //Refino/Remuevo creando el nuevo grid
           int ref,rem;
           ref=0;   rem=0;
            refinaIF(raiz,data,N,min_nivel_ref,max_nivel_ref, thetar,thetac,ref,rem);
            printf("Refinados = %d , Removidos = %d,  newN =%d\n", ref,rem,N+3*ref-3*rem);

         //Aplico la restriccion de suavidad
           for(int pp=0;pp<5;pp++)
           {
            int restri;  restri=0;
            restriccion_suavidad2(raiz,restri,max_nivel_ref);
           }

          //En el nuevo grid interpolo con respecto de lambda
            int N2,ni2;
            double *x2,*y2;
            double *H;
            tree   **data2;

            clean_frontera(raiz);   marca_frontera(raiz);
           // ajusta_fronteras(raiz,xmin,xmax,ymin,ymax);

            N2=0;  cont_hojas(raiz,N2);

             x2       = make_dvector(N2); 
             y2       = make_dvector(N2);
             H        = make_dvector(N2+6);
             data2 = (tree **)malloc(N2*sizeof(tree));   

            ni2=0; N2=0;
            get_datos_interior(raiz,x2,y2,data2,ni2);
            N2=ni2;
            get_datos_frontera(raiz,x2,y2,data2,N2);


          //Interpolo la solucion en  el nuevo grid para encontrar la lambda
              fill(H,N2+6,0.0);
              for(int i=0;i<N2;i++)  
                        H[i] =  interpola(x2[i],y2[i],x,y,lambda,0,N);


           //Calculo el error cuadratico medio y el error maximo
            for(int i=0;i<N;i++)    sol_analitica[i] = u_t(x[i],y[i],t);

            reconstruye(x,y,lambda,0,N,sol_numerica);

           e_rms = calc_erms(sol_analitica,sol_numerica,N);
           e_inf = calc_einf(sol_analitica,sol_numerica,N);
  
           for(int i=0;i<N;i++)     
                  sol_analitica_numerica[i] = sol_analitica[i]-sol_numerica[i];     
    
              save_gnu_solucion("numericat",x,y,sol_numerica,N);  
              save_gnu_solucion("realt",x,y,sol_analitica,N);     
              save_gnu_solucion("realnumericat",x,y,sol_analitica_numerica,N);      

              printf("t=%f, e_rms       = %f\n",t,e_rms);
              printf("t=%f, e_inf       = %f\n",t,e_inf);



          //Libero la memoria dinamica anterior
            free_dvector(x);
            free_dvector(y);
            free_dvector(lambda);
            free_ivector(pivot_index);
            free_dvector(f);
            free_dmatrix(P,N+6);
            free_dmatrix(M,N+6);

            free_dvector(sol_numerica);
            free_dvector(sol_analitica);
            free_dvector(sol_analitica_numerica);

         //Creo la nueva memoria dinamica
              N = N2;
             ni  = ni2;
              x = x2; 
              y = y2;
             lambda = make_dvector(N+6);
             f              = make_dvector(N+6);
            M             = make_dmatrix(N+6,N+6);
            P             = make_dmatrix(N+6,N+6);
            pivot_index     = make_ivector(N+6);

            sol_analitica                   = make_dvector(N);
           sol_numerica                  = make_dvector(N);
           sol_analitica_numerica = make_dvector(N);   


          //Creo las nuevas matrices 
                   crea_P(x2,y2, N2,stept,P);
                   crea_M(x2,y2,ni2, N2,stept,M);

         //Factorizo la matriz-M
                 Factor(M,N+6,&cond,pivot_index);

                 printf("num.condi   = %e\n",cond);  

         //Determino la nueva lambda
           fill(f,N+6,0.0);
           copy(H,N,f);   free_dvector(H);


           copy(f,N+6,lambda);
           Solve(M,N+6,pivot_index,lambda);

            free(data);
            data = data2;   

       }//if(maxi_elocal>thetar)
     } 
  }
  
  printf("\n");
  
//Calculo el error cuadratico medio y el error maximo
//sobre el mismo grid
  for(int i=0;i<N;i++)
    sol_analitica[i] = u_t(x[i],y[i],t);

  reconstruye(x,y,lambda,0,N,sol_numerica);

  e_rms = calc_erms(sol_analitica,sol_numerica,N);
  e_inf = calc_einf(sol_analitica,sol_numerica,N);
  
  for(int i=0;i<N;i++)     
     sol_analitica_numerica[i] = sol_analitica[i]-sol_numerica[i];     
    
  save_gnu_solucion("numericat",x,y,sol_numerica,N);  
  save_gnu_solucion("realt",x,y,sol_analitica,N);     
  save_gnu_solucion("realnumericat",x,y,sol_analitica_numerica,N);      

 printf("-----------------------------------------------------\n");     
 printf("Advection - Implicit - RLN - TPS\n");
 printf("thetar      = %f\n",thetar);   
 printf("thetac      = %f\n",thetac);    
printf("sigma      = %f\n",SIGMA);
 printf("stept       = %f\n",stept);   
 printf("tmax        = %f\n",tmax);    
 printf("t=%f, e_rms       = %f\n",t,e_rms);
 printf("t=%f, e_inf       = %f\n",t,e_inf);
 printf("-----------------------------------------------------\n");     
 

return 0;
}
/*

Con el metodo implicito y placa delgada para un problema dependiente
del tiempo  
                                Ut + Ux + Uy =0
se codifico la remocion e insercion de nodos.
Como resultado positivo, es que si sigue la forma de la gaussianda la
concentracion local de los nodos.
Como resultado negativo el centro de la gaussiana no corresponden al
valor analitico, de hecho se sobreestima. Hasta el momento no se porque pasa.


Source:     advection_imp_adap_tps.cpp
                  Pura Conveccion en 2-d  (una gaussiana desplazandose)
                  Metodo implicito 
                  Remocion e Insercion de Nodos  
                  Nucleo de placa delgada + Pol de grado 6 

Compile:
                       c++ advection_imp_adap_tps.cpp -lm

Example:

                       ./a.out 4 0.02 0.001 0.001 0.001 0.2

Output;
-----------------------------------------------------
Advection - Implicit - RLN - TPS
thetar      = 0.020000
thetac      = 0.001000
sigma      = 0.001000
stept       = 0.001000
tmax        = 0.200000
t=0.200000, e_rms       = 0.335653
t=0.200000, e_inf       = 0.926692
-----------------------------------------------------

*/



